"""
mngp.py
=======

Fully Bayesian Gaussian Process model that incorporates measured noise.

Created by Maxim Ziatdinov (email: maxim.ziatdinov@gmail.com)
"""

from typing import Callable, Dict, Optional, Tuple, Type, Union

import jax
import jaxlib
import jax.numpy as jnp
import jax.random as jra
import numpyro
import numpyro.distributions as dist
from numpyro.infer import MCMC, NUTS, init_to_median


from .gp import ExactGP
from .vigp import viGP
from .linreg import LinReg
from ..utils import get_keys

kernel_fn_type = Callable[[jnp.ndarray, jnp.ndarray, Dict[str, jnp.ndarray], jnp.ndarray], jnp.ndarray]


class MeasuredNoiseGP(ExactGP):
    """
    Gaussian Process model that incorporates measured noise.
    This class extends the ExactGP model by allowing the inclusion of measured noise variances 
    in the GP framework. Unlike standard GP models where noise is typically inferred, this model 
    uses noise values obtained from repeated measurements at the same input points.

    Args:
        input_dim:
            Number of input dimensions
        kernel:
            Kernel function ('RBF', 'Matern', 'Periodic', or custom function)
        mean_fn:
            Optional deterministic mean function (use 'mean_fn_priors' to make it probabilistic)
        kernel_prior:
            Optional custom priors over kernel hyperparameters. Use it when passing your custom kernel.
        mean_fn_prior:
            Optional priors over mean function parameters
        lengthscale_prior_dist:
            Optional custom prior distribution over kernel lengthscale. Defaults to LogNormal(0, 1).
    
    Examples:

        >>> # Get random number generator keys for training and prediction
        >>> key1, key2 = gpax.utils.get_keys()
        >>> # Initialize model
        >>> gp_model = gpax.MeasuredNoiseGP(input_dim=1, kernel='Matern')
        >>> # Run HMC to obtain posterior samples for the GP model parameters
        >>> gp_model.fit(key1, X, y_mean, noise)  # X, y_mean, and noise have dimensions (n, 1), (n,), and (n,)
        >>> # Make a prediction on new inputs by extrapolating noise variance with either linear regression or gaussian process
        >>> y_pred, y_samples = gp_model.predict(key2, X_new, noise_prediction_method='linreg')
    """
    def __init__(self,
                 input_dim: int,
                 kernel: Union[str, kernel_fn_type],
                 mean_fn: Optional[Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]] = None,
                 kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 mean_fn_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 lengthscale_prior_dist: Optional[dist.Distribution] = None
                 ) -> None:
        args = (input_dim, kernel, mean_fn, kernel_prior, mean_fn_prior, None, None, lengthscale_prior_dist)
        super(MeasuredNoiseGP, self).__init__(*args)
        self.measured_noise = None
        self.noise_predicted = None

    def model(self, X: jnp.ndarray, y: jnp.ndarray = None, measured_noise: jnp.ndarray = None, **kwargs) -> None:
        """GP model that accepts measured noise"""
        # Initialize mean function at zeros
        f_loc = jnp.zeros(X.shape[0])
        # Sample kernel parameters
        if self.kernel_prior:
            kernel_params = self.kernel_prior()
        else:
            kernel_params = self._sample_kernel_params()
        # Since we provide a measured noise, we don't infer it
        noise = numpyro.deterministic("noise", jnp.array(0.0))
        # Add mean function (if any)
        if self.mean_fn is not None:
            args = [X]
            if self.mean_fn_prior is not None:
                args += [self.mean_fn_prior()]
            f_loc += self.mean_fn(*args).squeeze()
        # compute kernel (with zero noise)
        k = self.kernel(X, X, kernel_params, 0, **kwargs)
        # Sample y according to the standard Gaussian process formula. Add measured noise to the covariance matrix
        numpyro.sample(
            "y",
            dist.MultivariateNormal(loc=f_loc, covariance_matrix=k+jnp.diag(measured_noise)),
            obs=y,
        )

    def fit(
        self,
        rng_key: jnp.array,
        X: jnp.ndarray,
        y: jnp.ndarray,
        measured_noise: jnp.ndarray,
        num_warmup: int = 2000,
        num_samples: int = 2000,
        num_chains: int = 1,
        chain_method: str = "sequential",
        progress_bar: bool = True,
        print_summary: bool = True,
        device: Type[jaxlib.xla_extension.Device] = None,
        **kwargs: float
    ) -> None:
        """
        Run Hamiltonian Monter Carlo to infer the GP parameters

        Args:
            rng_key: random number generator key
            X: 2D feature vector
            y: 1D target vector
            measured_noise: 1D vector with measured noise
            num_warmup: number of HMC warmup states
            num_samples: number of HMC samples
            num_chains: number of HMC chains
            chain_method: 'sequential', 'parallel' or 'vectorized'
            progress_bar: show progress bar
            print_summary: print summary at the end of sampling
            device:
                optionally specify a cpu or gpu device on which to run the inference;
                e.g., ``device=jax.devices("cpu")[0]``
            **jitter:
                Small positive term added to the diagonal part of a covariance
                matrix for numerical stability (Default: 1e-6)
        """
        X, y = self._set_data(X, y)
        if device:
            X = jax.device_put(X, device)
            y = jax.device_put(y, device)
        self.X_train = X
        self.y_train = y
        self.measured_noise = measured_noise

        init_strategy = init_to_median(num_samples=10)
        kernel = NUTS(self.model, init_strategy=init_strategy)
        self.mcmc = MCMC(
            kernel,
            num_warmup=num_warmup,
            num_samples=num_samples,
            num_chains=num_chains,
            chain_method=chain_method,
            progress_bar=progress_bar,
            jit_model_args=False,
        )
        self.mcmc.run(rng_key, X, y, measured_noise, **kwargs)
        if print_summary:
            self._print_summary()

    def _predict(
        self,
        rng_key: jnp.ndarray,
        X_new: jnp.ndarray,
        params: Dict[str, jnp.ndarray],
        noise_predicted: jnp.ndarray,
        n: int,
        noiseless: bool = False,
        **kwargs: float) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """Prediction with a single sample of GP parameters"""

        def sigma_sample(rng_key, K, X_new_shape):
            sig = jnp.sqrt(jnp.clip(jnp.diag(K), a_min=0.0))
            return sig * jra.normal(rng_key, X_new_shape[:1])
        
        # Get the predictive mean and covariance
        y_mean, K = self.get_mvn_posterior(X_new, params, noiseless, **kwargs)
        # Add predicted noise to K's diagonal
        K += jnp.diag(noise_predicted)
        # Draw samples from the posterior predictive for a given set of parameters
        rng_keys = jra.split(rng_key, n)
        sig = jax.vmap(sigma_sample, in_axes=(0, None, None))(rng_keys, K, X_new.shape)
        y_sampled = y_mean + sig
        return y_mean, y_sampled
    
    def predict(
        self,
        rng_key: jnp.ndarray,
        X_new: jnp.ndarray,
        samples: Optional[Dict[str, jnp.ndarray]] = None,
        n: int = 1,
        filter_nans: bool = False,
        noiseless: bool = True,
        device: Type[jaxlib.xla_extension.Device] = None,
        noise_prediction_method: str = 'linreg',
        **kwargs: float
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Make prediction at X_new points using posterior samples for GP parameters

        Args:
            rng_key: random number generator key
            X_new: new inputs with *(number of points, number of features)* dimensions
            samples: optional (different) samples with GP parameters
            n: number of samples from Multivariate Normal posterior for each HMC sample with GP parameters
            filter_nans: filter out samples containing NaN values (if any)
            noiseless:
                Noise-free prediction. It is set to False by default as new/unseen data is assumed
                to follow the same distribution as the training data. Hence, since we introduce a model noise
                by default for the training data, we also want to include that noise in our prediction.
            device:
                optionally specify a cpu or gpu device on which to make a prediction;
                e.g., ```device=jax.devices("gpu")[0]```
            noise_prediction_method:
                Method for extrapolating noise variance to new/test data.
                Choose between 'linreg' and 'gpreg'. Defaults to 'linreg'.
            **jitter:
                Small positive term added to the diagonal part of a covariance
                matrix for numerical stability (Default: 1e-6)

        Returns
            Center of the mass of sampled means and all the sampled predictions
        """
        if noise_prediction_method not in ["linreg", "gpreg"]:
            raise NotImplementedError(
                "For noise prediction method, select between 'linreg' and 'gpreg'")
        noise_pred_fn = self.linreg if noise_prediction_method == "linreg" else self.gpreg
        X_new = self._set_data(X_new)
        # Predict noise for X_new
        if self.noise_predicted is not None:
            noise_predicted = self.noise_predicted
        else:
            noise_predicted = noise_pred_fn(self.X_train, self.measured_noise, X_new, **kwargs)
            self.noise_predicted = noise_predicted
        if samples is None:
            samples = self.get_samples(chain_dim=False)
        if device:
            self._set_training_data(device=device)
            X_new = jax.device_put(X_new, device)
            samples = jax.device_put(samples, device)
        num_samples = len(next(iter(samples.values())))
        vmap_args = (jra.split(rng_key, num_samples), samples)
        predictive = jax.vmap(lambda prms: self._predict(prms[0], X_new, prms[1], noise_predicted, n, noiseless, **kwargs))
        y_means, y_sampled = predictive(vmap_args)
        if filter_nans:
            y_sampled_ = [y_i for y_i in y_sampled if not jnp.isnan(y_i).any()]
            y_sampled = jnp.array(y_sampled_)
        return y_means.mean(0), y_sampled
    
    def linreg(self, x, y, x_new, **kwargs):
        lreg = LinReg()
        lreg.train(x, y, **kwargs)
        return lreg.predict(x_new)
    
    def gpreg(self, x, y, x_new, **kwargs):
        keys = get_keys()
        vigp = viGP(self.kernel_dim, 'RBF', **kwargs)
        vigp.fit(keys[0], x, y, progress_bar=False, print_summary=False, **kwargs)
        return vigp.predict(keys[1], x_new, noiseless=True)[0]


import sys
import pytest
import numpy as onp
import jax.numpy as jnp
import jax
import numpyro
import numpyro.distributions as dist
from numpy.testing import assert_equal, assert_array_equal, assert_

sys.path.insert(0, "../gpax/")

from gpax.models.hskgp import VarNoiseGP
from gpax.utils import get_keys


def get_dummy_data(unsqueeze=False):
    X = onp.linspace(1, 2, 8) + 0.1 * onp.random.randn(8,)
    y = (10 * X**2)
    if unsqueeze:
        X = X[:, None]
    return jnp.array(X), jnp.array(y)


def noise_fn(x, params):
    return params["a"] + params["b"]*x


def noise_fn_prior():
    a = numpyro.sample("a", dist.Normal(0, 1))
    b = numpyro.sample("b", dist.Normal(0, 1))
    return {"a": a, "b": b}


@pytest.mark.parametrize("noise_kernel", ['RBF', 'Matern'])
def test_fit(noise_kernel):
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    m = VarNoiseGP(1, 'RBF', noise_kernel=noise_kernel)
    m.fit(rng_key, X, y, num_warmup=10, num_samples=10)
    assert m.mcmc is not None


def test_fit_with_custom_noise_lscale():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    m = VarNoiseGP(1, 'RBF', noise_lengthscale_prior_dist=dist.HalfNormal(1))
    m.fit(rng_key, X, y, num_warmup=10, num_samples=10)
    assert m.mcmc is not None


def test_fit_with_noise_mean_fn():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    m = VarNoiseGP(1, 'RBF', noise_mean_fn=noise_fn, noise_mean_fn_prior=noise_fn_prior)
    m.fit(rng_key, X, y, num_warmup=10, num_samples=10)
    assert m.mcmc is not None


def test_fit_with_noise_and_regular_mean_fn():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    m = VarNoiseGP(1, 'RBF', mean_fn = lambda x: 8*x**2,
                   noise_mean_fn=noise_fn, noise_mean_fn_prior=noise_fn_prior)
    m.fit(rng_key, X, y, num_warmup=10, num_samples=10)
    assert m.mcmc is not None


def test_get_mvn_posterior():
    X, y = get_dummy_data(unsqueeze=True)
    X_test, _ = get_dummy_data(unsqueeze=True)
    params = {"k_length": jnp.array([1.0]),
              "k_scale": jnp.array(1.0),
              "noise": jnp.array(0.1),
              "k_noise_length": jnp.array(0.5),
              "k_noise_scale": jnp.array(1.0),
              "log_var": jnp.ones(len(X))}
    m = VarNoiseGP(1, 'RBF', noise_kernel='RBF')
    m.X_train = X
    m.y_train = y
    mean, cov = m.get_mvn_posterior(X_test, params)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(cov, jnp.ndarray)
    assert_equal(mean.shape, (X_test.shape[0],))
    assert_equal(cov.shape, (X_test.shape[0], X_test.shape[0]))


def test_get_mvn_posterior_with_mean_fn():
    X, y = get_dummy_data(unsqueeze=True)
    X_test, _ = get_dummy_data(unsqueeze=True)
    params = {"k_length": jnp.array([1.0]),
              "k_scale": jnp.array(1.0),
              "noise": jnp.array(0.1),
              "k_noise_length": jnp.array(0.5),
              "k_noise_scale": jnp.array(1.0),
              "log_var": jnp.ones(len(X)),
              "a": jnp.array(1.0),
              "b": jnp.array(1.0)
              }
    m = VarNoiseGP(1, 'RBF', noise_kernel='RBF', noise_mean_fn=noise_fn, noise_mean_fn_prior=noise_fn_prior)
    m.X_train = X
    m.y_train = y
    mean, cov = m.get_mvn_posterior(X_test, params)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(cov, jnp.ndarray)
    assert_equal(mean.shape, (X_test.shape[0],))
    assert_equal(cov.shape, (X_test.shape[0], X_test.shape[0]))


def test_get_mvn_posterior_with_noise_and_regular_mean_fn():
    X, y = get_dummy_data(unsqueeze=True)
    X_test, _ = get_dummy_data(unsqueeze=True)
    params = {"k_length": jnp.array([1.0]),
              "k_scale": jnp.array(1.0),
              "noise": jnp.array(0.1),
              "k_noise_length": jnp.array(0.5),
              "k_noise_scale": jnp.array(1.0),
              "log_var": jnp.ones(len(X)),
              "a": jnp.array(1.0),
              "b": jnp.array(1.0)
              }
    m = VarNoiseGP(1, 'RBF', noise_kernel='RBF',
                   mean_fn = lambda x: 8*x**2,
                   noise_mean_fn=noise_fn, noise_mean_fn_prior=noise_fn_prior)
    m.X_train = X
    m.y_train = y
    mean, cov = m.get_mvn_posterior(X_test, params)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(cov, jnp.ndarray)
    assert_equal(mean.shape, (X_test.shape[0],))
    assert_equal(cov.shape, (X_test.shape[0], X_test.shape[0]))


def test_get_noise_samples():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    m = VarNoiseGP(1, 'RBF')
    m.fit(rng_key, X, y, num_warmup=10, num_samples=10)
    noise = m.get_data_var_samples()
    assert_(isinstance(noise, jnp.ndarray))


def test_get_noise_samples_with_mean_fn():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    m = VarNoiseGP(1, 'RBF', noise_mean_fn=noise_fn, noise_mean_fn_prior=noise_fn_prior)
    m.fit(rng_key, X, y, num_warmup=10, num_samples=10)
    noise = m.get_data_var_samples()
    assert_(isinstance(noise, jnp.ndarray))



import sys
import pytest
import numpy as onp
import jax.numpy as jnp
from numpy.testing import assert_

sys.path.insert(0, "../gpax/")

from gpax.models.gp import ExactGP
from gpax.acquisition.optimize import optimize_acq
from gpax.acquisition.acquisition import UCB, EI
from gpax.utils import get_keys


def get_inputs():
    X = onp.random.uniform(-2, 2, size=(4,))
    y = X**3
    return X, y


@pytest.mark.parametrize("acq_fn", [UCB, EI])
def test_optimize_acq(acq_fn):
    lower_bound = -2.0
    upper_bound = 2.0
    num_initial_guesses = 3
    key1, key2 = get_keys()
    X, y = get_inputs()
    model = ExactGP(1, 'RBF')
    model.fit(key1, X, y, num_warmup=50, num_samples=50)
    x_next = optimize_acq(
        key2, model, acq_fn, num_initial_guesses, lower_bound, upper_bound)
    assert_(isinstance(x_next, jnp.ndarray))
    


    


import sys
import pytest
import numpy as onp
import jax.numpy as jnp
import jax
import numpyro
import numpyro.distributions as dist
from numpy.testing import assert_equal, assert_array_equal, assert_

sys.path.insert(0, "../gpax/")

from gpax.models.uigp import UIGP
from gpax.utils import get_keys


def get_dummy_data():
    X = onp.linspace(1, 2, 8) + 0.1 * onp.random.randn(8,)
    X_prime = onp.random.normal(X, 0.1)
    y = (10 * X_prime**2)
    return jnp.array(X_prime), jnp.array(y)


@pytest.mark.parametrize("n_features", [1, 5])
def test_sample_x(n_features):
    X = onp.random.randn(32, n_features)
    m = UIGP(n_features, 'RBF')
    with numpyro.handlers.seed(rng_seed=0):
        X_prime = m._sample_x(X)
    assert_(isinstance(X_prime, jnp.ndarray))
    assert_(X_prime.shape[-1], n_features)


def test_fit():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    m = UIGP(1, 'RBF')
    m.fit(rng_key, X, y, num_warmup=10, num_samples=10)
    assert_(m.mcmc is not None)


def test_fit_with_custom_sigma_x_prior():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    m = UIGP(1, 'RBF', sigma_x_prior_dist=dist.HalfNormal(0.55))
    m.fit(rng_key, X, y, num_warmup=10, num_samples=10)
    assert_(m.mcmc is not None)


def test_get_mvn_posterior():
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    X = X[:, None]
    X_test = X_test[:, None]
    params = {"k_length": jnp.array([1.0]),
              "k_scale": jnp.array(1.0),
              "noise": jnp.array(0.1),
              "k_noise_length": jnp.array(0.5),
              "sigma_x": jnp.array(0.3),
              "X_prime": jnp.array(X + 0.1)
              }
    m = UIGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    mean, cov = m.get_mvn_posterior(X_test, params)
    assert_(isinstance(mean, jnp.ndarray))
    assert_(isinstance(cov, jnp.ndarray))
    assert_equal(mean.shape, (X_test.shape[0],))
    assert_equal(cov.shape, (X_test.shape[0], X_test.shape[0]))


@pytest.mark.parametrize("noiseless", [True, False])
def test_predict_single_sample(noiseless):
    key = get_keys()[0]
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    X = X[:, None]
    X_test = X_test[:, None]
    params = {"k_length": jnp.array([1.0]),
              "k_scale": jnp.array(1.0),
              "noise": jnp.array(0.1),
              "sigma_x": jnp.array(0.3),
              "X_prime": jnp.array(X + 0.1)
              }
    m = UIGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    mean, sample = m._predict(key, X_test, params, 5, noiseless)
    assert_(isinstance(mean, jnp.ndarray))
    assert_(isinstance(sample, jnp.ndarray))
    assert_equal(mean.shape, (X_test.shape[0],))
    assert_equal(sample.shape, (5, X_test.shape[0]))



from .acquisition import UCB, EI, POI, UE, Thompson, KG
from .batch_acquisition import qEI, qPOI, qUCB, qKG
from .optimize import optimize_acq

__all__ = ["UCB", "EI", "POI", "UE", "KG", "Thompson", "qEI", "qPOI", "qUCB", "qKG", "optimize_acq"]



"""
priors.py
=========

Utility functions for setting priors

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

import inspect
import re

from typing import Union, Dict, Type, List, Callable, Optional

import numpyro
import jax
import jax.numpy as jnp

from ..kernels.kernels import square_scaled_distance, add_jitter, _sqrt


def place_normal_prior(param_name: str, loc: float = 0.0, scale: float = 1.0):
    """
    Samples a value from a normal distribution with the specified mean (loc) and standard deviation (scale),
    and assigns it to a named random variable in the probabilistic model. Can be useful for defining prior mean functions
    in structured Gaussian processes.
    """
    return numpyro.sample(param_name, normal_dist(loc, scale))


def place_lognormal_prior(param_name: str, loc: float = 0.0, scale: float = 1.0):
    """
    Samples a value from a log-normal distribution with the specified mean (loc) and standard deviation (scale),
    and assigns it to a named random variable in the probabilistic model. Can be useful for defining prior mean functions
    in structured Gaussian processes.
    """
    return numpyro.sample(param_name, lognormal_dist(loc, scale))


def place_halfnormal_prior(param_name: str, scale: float = 1.0):
    """
    Samples a value from a half-normal distribution with the specified standard deviation (scale),
    and assigns it to a named random variable in the probabilistic model. Can be useful for defining prior mean functions
    in structured Gaussian processes.
    """
    return numpyro.sample(param_name, halfnormal_dist(scale))


def place_uniform_prior(param_name: str,
                        low: float = None,
                        high: float = None,
                        X: jnp.ndarray = None):
    """
    Samples a value from a uniform distribution with the specified low and high values,
    and assigns it to a named random variable in the probabilistic model. Can be useful for defining prior mean functions
    in structured Gaussian processes.
    """
    d = uniform_dist(low, high, X)
    return numpyro.sample(param_name, d)


def place_gamma_prior(param_name: str,
                      c: float = None,
                      r: float = None,
                      X: jnp.ndarray = None):
    """
    Samples a value from a uniform distribution with the specified concentration (c) and rate (r) values,
    and assigns it to a named random variable in the probabilistic model. Can be useful for defining prior mean functions
    in structured Gaussian processes.
    """
    d = gamma_dist(c, r, X)
    return numpyro.sample(param_name, d)


def normal_dist(loc: float = None, scale: float = None
                ) -> numpyro.distributions.Distribution:
    """
    Generate a Normal distribution based on provided center (loc) and standard deviation (scale) parameters.
    I neithere are provided, uses 0 and 1 by default.
    """
    loc = loc if loc is not None else 0.0
    scale = scale if scale is not None else 1.0
    return numpyro.distributions.Normal(loc, scale)


def lognormal_dist(loc: float = None, scale: float = None) -> numpyro.distributions.Distribution:
    """
    Generate a LogNormal distribution based on provided center (loc) and standard deviation (scale) parameters.
    I neithere are provided, uses 0 and 1 by default.
    """
    loc = loc if loc is not None else 0.0
    scale = scale if scale is not None else 1.0
    return numpyro.distributions.LogNormal(loc, scale)


def halfnormal_dist(scale: float = None) -> numpyro.distributions.Distribution:
    """
    Generate a half-normal distribution based on provided standard deviation (scale).
    If none is provided, uses 1.0 by default.
    """
    scale = scale if scale is not None else 1.0
    return numpyro.distributions.HalfNormal(scale)


def gamma_dist(c: float = None,
               r: float = None,
               input_vec: jnp.ndarray = None
               ) -> numpyro.distributions.Distribution:
    """
    Generate a Gamma distribution based on provided shape (c) and rate (r) parameters. If the shape (c) is not provided,
    it attempts to infer it using the range of the input vector divided by 2. The rate parameter defaults to 1.0 if not provided.
    """
    if c is None:
        if input_vec is not None:
            c = (input_vec.max() - input_vec.min()) / 2
        else:
            raise ValueError("Provide either c or an input array")
    if r is None:
        r = 1.0
    return numpyro.distributions.Gamma(c, r)


def uniform_dist(low: float = None,
                 high: float = None,
                 input_vec: jnp.ndarray = None
                 ) -> numpyro.distributions.Distribution:
    """
    Generate a Uniform distribution based on provided low and high bounds. If one of the bounds is not provided,
    it attempts to infer the missing bound(s) using the minimum or maximum value from the input vector.
    """
    if (low is None or high is None) and input_vec is None:
        raise ValueError(
            "If 'low' or 'high' is not provided, an input array must be provided.")
    low = low if low is not None else input_vec.min()
    high = high if high is not None else input_vec.max()

    return numpyro.distributions.Uniform(low, high)


def set_fn(func: Callable) -> Callable:
    """
    Transforms the given deterministic function to use a params dictionary
    for its parameters, excluding the first one (assumed to be the dependent variable).

    Args:
    - func (Callable): The deterministic function to be transformed.

    Returns:
    - Callable: The transformed function where parameters are accessed
                from a `params` dictionary.
    """
    # Extract parameter names excluding the first one (assumed to be the dependent variable)
    params_names = list(inspect.signature(func).parameters.keys())[1:]

    # Create the transformed function definition
    transformed_code = f"def {func.__name__}(x, params):\n"

    # Retrieve the source code of the function and indent it to be a valid function body
    source = inspect.getsource(func).split("\n", 1)[1]
    source = "    " + source.replace("\n", "\n    ")

    # Replace each parameter name with its dictionary lookup using regex
    for name in params_names:
        source = re.sub(rf'\b{name}\b', f'params["{name}"]', source)

    # Combine to get the full source
    transformed_code += source

    # Define the transformed function in the local namespace
    local_namespace = {}
    exec(transformed_code, globals(), local_namespace)

    # Return the transformed function
    return local_namespace[func.__name__]


def set_kernel_fn(func: Callable,
                  independent_vars: List[str] = ["X", "Z"],
                  jit_decorator: bool = True,
                  docstring: Optional[str] = None) -> Callable:
    """
    Transforms the given kernel function to use a params dictionary for its hyperparameters.
    The resultant function will always add jitter before returning the computed kernel.

    Args:
        func (Callable): The kernel function to be transformed.
        independent_vars (List[str], optional): List of independent variable names in the function. Defaults to ["X", "Z"].
        jit_decorator (bool, optional): @jax.jit decorator to be applied to the transformed function. Defaults to True.
        docstring (Optional[str], optional): Docstring to be added to the transformed function. Defaults to None.

    Returns:
        Callable: The transformed kernel function where hyperparameters are accessed from a `params` dictionary.
    """

    # Extract parameter names excluding the independent variables
    params_names = [k for k, v in inspect.signature(func).parameters.items() if v.default == v.empty]
    for var in independent_vars:
        params_names.remove(var)

    transformed_code = ""
    if jit_decorator:
        transformed_code += "@jit" + "\n"

    additional_args = "noise: int = 0, jitter: float = 1e-6, **kwargs"
    transformed_code += f"def {func.__name__}({', '.join(independent_vars)}, params: Dict[str, jnp.ndarray], {additional_args}):\n"

    if docstring:
        transformed_code += '    """' + docstring + '"""\n'

    source = inspect.getsource(func).split("\n", 1)[1]
    lines = source.split("\n")

    for idx, line in enumerate(lines):
        # Convert all parameter names to their dictionary lookup throughout the function body
        for name in params_names:
            lines[idx] = re.sub(rf'\b{name}\b', f'params["{name}"]', lines[idx])

    # Combine lines back and then split again by return
    modified_source = '\n'.join(lines)
    pre_return, return_statement = modified_source.split('return', 1)

    # Append custom jitter code
    custom_code = f"    {pre_return.strip()}\n    k = {return_statement.strip()}\n"
    custom_code += """
    if X.shape == Z.shape:
        k += (noise + jitter) * jnp.eye(X.shape[0])
    return k
    """

    transformed_code += custom_code

    local_namespace = {"jit": jax.jit}
    exec(transformed_code, globals(), local_namespace)

    return local_namespace[func.__name__]


def _set_noise_kernel_fn(func: Callable) -> Callable:
    """
    Modifies the GPax kernel function to append "_noise" after "k" in dictionary keys it accesses.

    Args:
        func (Callable): Original function.

    Returns:
        Callable: Modified function.
    """

    # Get the source code of the function
    source = inspect.getsource(func)

    # Split the source into decorators, definition, and body
    decorators_and_def, body = source.split("\n", 1)

    # Replace all occurrences of params["k with params["k_noise in the body
    modified_body = re.sub(r'params\["k', 'params["k_noise', body)

    # Combine decorators, definition, and modified body
    modified_source = f"{decorators_and_def}\n{modified_body}"

    # Define local namespace including the jit decorator
    local_namespace = {"jit": jax.jit}

    # Execute the modified source to redefine the function in the provided namespace
    exec(modified_source, globals(), local_namespace)

    # Return the modified function
    return local_namespace[func.__name__]


def auto_priors(func: Callable, params_begin_with: int, dist_type: str = 'normal', loc: float = 0.0, scale: float = 1.0) -> Callable:
    """
    Generates a function that, when invoked, samples from normal or log-normal distributions
    for each parameter of the given deterministic function, except the first one.

    Args:
        func (Callable): The deterministic function for which to set normal or log-normal priors.
        params_begin_with (int): Parameters to account for start from this number.
        loc (float, optional): Mean of the normal or log-normal distribution. Defaults to 0.0.
        scale (float, optional): Standard deviation of the normal or log-normal distribution. Defaults to 1.0.

    Returns:
        A function that, when invoked, returns a dictionary of sampled values
        from normal or log-normal distributions for each parameter of the original function.
    """
    place_prior = place_lognormal_prior if dist_type == 'lognormal' else place_normal_prior

    # Get the names of the parameters of the function excluding the first one (dependent variable)
    params_names = list(inspect.signature(func).parameters.keys())[params_begin_with:]

    def sample_priors() -> Dict[str, Union[float, Type[Callable]]]:
        # Return a dictionary with normal priors for each parameter
        return {name: place_prior(name, loc, scale) for name in params_names}

    return sample_priors


def auto_normal_priors(func: Callable, loc: float = 0.0, scale: float = 1.0) -> Callable:
    """
    Places normal priors over function parameters.

    Args:
        func (Callable): The deterministic function for which to set normal priors.
        loc (float, optional): Mean of the normal distribution. Defaults to 0.0.
        scale (float, optional): Standard deviation of the normal distribution. Defaults to 1.0.

    Returns:
        A function that, when invoked, returns a dictionary of sampled values
        from normal distributions for each parameter of the original function.
    """
    return auto_priors(func, 1, 'normal', loc, scale)


def auto_lognormal_priors(func: Callable, loc: float = 0.0, scale: float = 1.0) -> Callable:
    """
    Places log-normal priors over function parameters.

    Args:
        func (Callable): The deterministic function for which to set log-normal priors.
        loc (float, optional): Mean of the log-normal distribution. Defaults to 0.0.
        scale (float, optional): Standard deviation of the log-normal distribution. Defaults to 1.0.

    Returns:
        A function that, when invoked, returns a dictionary of sampled values
        from log-normal distributions for each parameter of the original function.
    """
    return auto_priors(func, 1, 'lognormal', loc, scale)


def auto_normal_kernel_priors(kernel_fn: Callable, loc: float = 0.0, scale: float = 1.0) -> Callable:
    """
    Places normal priors over the kernel parameters.

    Args:
        func (Callable): The deterministic kernel function for which to set normal priors.
        loc (float, optional): Mean of the normal distribution. Defaults to 0.0.
        scale (float, optional): Standard deviation of the normal distribution. Defaults to 1.0.

    Returns:
        A function that, when invoked, returns a dictionary of sampled values
        from normal distributions for each parameter of the original kernel function.
    """
    return auto_priors(kernel_fn, 2, 'normal', loc, scale)


def auto_lognormal_kernel_priors(kernel_fn: Callable, loc: float = 0.0, scale: float = 1.0) -> Callable:
    """
    Places log-normal priors over the kernel parameters.

    Args:
        func (Callable): The deterministic kernel function for which to set log-normal priors.
        loc (float, optional): Mean of the log-normal distribution. Defaults to 0.0.
        scale (float, optional): Standard deviation of the log-normal distribution. Defaults to 1.0.

    Returns:
        A function that, when invoked, returns a dictionary of sampled values
        from log-normal distributions for each parameter of the original kernel function.
    """
    return auto_priors(kernel_fn, 2, 'lognormal', loc, scale)



from .gp import ExactGP
from .vgp import vExactGP
from .vigp import viGP
from .hskgp import VarNoiseGP
from .spm import sPM
from .ibnn import iBNN
from .vi_ibnn import vi_iBNN
from .dkl import DKL
from .vidkl import viDKL
from .vi_mtdkl import viMTDKL
from .mtgp import MultiTaskGP
from .corgp import CoregGP
from .uigp import UIGP
from .mngp import MeasuredNoiseGP
from .linreg import LinReg

__all__ = [
    "ExactGP",
    "vExactGP",
    "viGP",
    "VarNoiseGP",
    "sPM",
    "iBNN",
    "vi_iBNN",
    "DKL",
    "viDKL",
    "viMTDKL",
    "MultiTaskGP",
    "CoregGP",
    "UIGP",
    "LinReg",
    "MeasuredNoiseGP"
]



import sys
import pytest
import numpy as onp
import jax.numpy as jnp
import jax
import numpyro
import numpyro.distributions as dist
from numpy.testing import assert_equal, assert_array_equal, assert_

sys.path.insert(0, "../gpax/")

from gpax.models.mngp import MeasuredNoiseGP
from gpax.utils import get_keys


def variable_noise(x):
    return 0.1 + 0.5 * x

def get_dummy_data():
    f = lambda x: onp.sin(x) * x
    X = onp.linspace(1, 2, 8)
    y_all = onp.array([f(x) + onp.random.normal(0, variable_noise(x), 10) for x in X])
    y = y_all.mean(1)
    measured_noise = y_all.var(1)
    return jnp.array(X), jnp.array(y), jnp.array(measured_noise)


def test_fit():
    rng_key = get_keys()[0]
    X, y, measured_noise = get_dummy_data()
    m = MeasuredNoiseGP(1, 'RBF')
    m.fit(rng_key, X, y, measured_noise, num_warmup=10, num_samples=10)
    assert m.mcmc is not None


def test_get_mvn_posterior():
    X, y, measured_noise = get_dummy_data()
    X_test, _, _ = get_dummy_data()
    X = X[:, None]
    X_test = X_test[:, None]
    params = {"k_length": jnp.array([1.0]),
              "k_scale": jnp.array(1.0),
              "noise": jnp.array(0.0),
              }
    m = MeasuredNoiseGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    m.measured_noise = measured_noise
    mean, cov = m.get_mvn_posterior(X_test, params)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(cov, jnp.ndarray)
    assert_equal(mean.shape, (X_test.shape[0],))
    assert_equal(cov.shape, (X_test.shape[0], X_test.shape[0]))


@pytest.mark.parametrize("n", [1, 5])
def test_predict_single_sample(n):
    key = get_keys()[0]
    X, y, measured_noise = get_dummy_data()
    X_test, _, _ = get_dummy_data()
    X = X[:, None]
    X_test = X_test[:, None]
    params = {"k_length": jnp.array([1.0]),
              "k_scale": jnp.array(1.0),
              "noise": jnp.array(0.0),
              }
    m = MeasuredNoiseGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    m.measured_noise = measured_noise
    noise_predicted = 0.5 * X_test
    mean, sample = m._predict(key, X_test, params, noise_predicted, n)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(sample, jnp.ndarray)
    assert_equal(mean.shape, (X_test.shape[0],))
    assert_equal(sample.shape, (n, X_test.shape[0]))
 

@pytest.mark.parametrize("noise_pred_fn", ['linreg', 'gpreg'])
def test_predict(noise_pred_fn):
    rng_keys = get_keys()
    X, y, measured_noise = get_dummy_data()
    X = X[:, None]
    X_test, _, _ = get_dummy_data()
    samples = {"k_length": jax.random.normal(rng_keys[0], shape=(100, 1)),
               "k_scale": jax.random.normal(rng_keys[0], shape=(100,)),
               "noise": jax.random.normal(rng_keys[0], shape=(100,))}
    m = MeasuredNoiseGP(1, 'RBF')
    m.X_train = X
    m.y_train = y
    m.measured_noise = measured_noise
    y_mean, y_sampled = m.predict(rng_keys[1], X_test, samples, noise_prediction_method=noise_pred_fn)
    assert isinstance(y_mean, jnp.ndarray)
    assert isinstance(y_sampled, jnp.ndarray)
    assert_equal(y_mean.shape, (X_test.shape[0],))
    assert_equal(y_sampled.shape, (100, 1, X_test.shape[0]))
    



import sys
import pytest
import jax.numpy as jnp
import numpyro
from numpy.testing import assert_equal, assert_

sys.path.insert(0, "../gpax/")

from gpax.utils import place_normal_prior, place_halfnormal_prior, place_uniform_prior, place_gamma_prior, place_lognormal_prior
from gpax.utils import uniform_dist, normal_dist, halfnormal_dist, lognormal_dist, gamma_dist
from gpax.utils import auto_lognormal_priors, auto_normal_priors, auto_lognormal_kernel_priors, auto_normal_kernel_priors, auto_priors
from gpax.utils import set_fn, set_kernel_fn, _set_noise_kernel_fn


def linear_kernel_test(X, Z, k_scale):
    # Dummy kernel functions for testing purposes
    return k_scale * jnp.dot(X, Z.T)


def rbf_test(X, Z, k_length, k_scale):
    # Dummy kernel functions for testing purposes
    scaled_X = X / k_length
    scaled_Z = Z / k_length
    X2 = (scaled_X ** 2).sum(1, keepdims=True)
    Z2 = (scaled_Z ** 2).sum(1, keepdims=True)
    XZ = jnp.matmul(scaled_X, scaled_Z.T)
    r2 = X2 - 2 * XZ + Z2.T

    return k_scale * jnp.exp(-0.5 * r2)


def sample_function(x, a, b):
    return a + b * x


@pytest.mark.parametrize("prior", [place_normal_prior, place_halfnormal_prior, place_lognormal_prior])
def test_normal_prior(prior):
    with numpyro.handlers.seed(rng_seed=1):
        sample = prior("a")
    assert_(isinstance(sample, jnp.ndarray))


def test_uniform_prior():
    with numpyro.handlers.seed(rng_seed=1):
        sample = place_uniform_prior("a", 0, 1)
    assert_(isinstance(sample, jnp.ndarray))


def test_gamma_prior():
    with numpyro.handlers.seed(rng_seed=1):
        sample = place_gamma_prior("a", 2, 2)
    assert_(isinstance(sample, jnp.ndarray))


def test_normal_prior_params():
    with numpyro.handlers.seed(rng_seed=1):
        with numpyro.handlers.trace() as tr:
            place_normal_prior("a", loc=0.5, scale=0.1)
    site = tr["a"]
    assert_(isinstance(site['fn'], numpyro.distributions.Normal))
    assert_equal(site['fn'].loc, 0.5)
    assert_equal(site['fn'].scale, 0.1)


def test_lognormal_prior_params():
    with numpyro.handlers.seed(rng_seed=1):
        with numpyro.handlers.trace() as tr:
            place_lognormal_prior("a", loc=0.5, scale=0.1)
    site = tr["a"]
    assert_(isinstance(site['fn'], numpyro.distributions.LogNormal))
    assert_equal(site['fn'].loc, 0.5)
    assert_equal(site['fn'].scale, 0.1)


def test_halfnormal_prior_params():
    with numpyro.handlers.seed(rng_seed=1):
        with numpyro.handlers.trace() as tr:
            place_halfnormal_prior("a", 0.1)
    site = tr["a"]
    assert_(isinstance(site['fn'], numpyro.distributions.HalfNormal))
    assert_equal(site['fn'].scale, 0.1)


def test_uniform_prior_params():
    with numpyro.handlers.seed(rng_seed=1):
        with numpyro.handlers.trace() as tr:
            place_uniform_prior("a", low=0.5, high=1.0)
    site = tr["a"]
    assert_(isinstance(site['fn'], numpyro.distributions.Uniform))
    assert_equal(site['fn'].low, 0.5)
    assert_equal(site['fn'].high, 1.0)


def test_gamma_prior_params():
    with numpyro.handlers.seed(rng_seed=1):
        with numpyro.handlers.trace() as tr:
            place_gamma_prior("a", c=2.0, r=1.0)
    site = tr["a"]
    assert_(isinstance(site['fn'], numpyro.distributions.Gamma))
    assert_equal(site['fn'].concentration, 2.0)
    assert_equal(site['fn'].rate, 1.0)


def test_get_uniform_dist():
    uniform_dist_ = uniform_dist(low=1.0, high=5.0)
    assert isinstance(uniform_dist_, numpyro.distributions.Uniform)
    assert uniform_dist_.low == 1.0
    assert uniform_dist_.high == 5.0


def test_get_uniform_dist_infer_params():
    uniform_dist_ = uniform_dist(input_vec=jnp.array([1.0, 2.0, 3.0, 4.0, 5.0]))
    assert uniform_dist_.low == 1.0
    assert uniform_dist_.high == 5.0


def test_get_gamma_dist():
    gamma_dist_ = gamma_dist(c=2.0, r=1.0)
    assert isinstance(gamma_dist_, numpyro.distributions.Gamma)
    assert gamma_dist_.concentration == 2.0
    assert gamma_dist_.rate == 1.0


def test_get_normal_dist():
    normal_dist_ = normal_dist(loc=2.0, scale=3.0)
    assert isinstance(normal_dist_, numpyro.distributions.Normal)
    assert normal_dist_.loc == 2.0
    assert normal_dist_.scale == 3.0


def test_get_lognormal_dist():
    lognormal_dist_ = lognormal_dist(loc=2.0, scale=3.0)
    assert isinstance(lognormal_dist_, numpyro.distributions.LogNormal)
    assert lognormal_dist_.loc == 2.0
    assert lognormal_dist_.scale == 3.0


def test_get_halfnormal_dist():
    halfnormal_dist_ = halfnormal_dist(scale=1.5)
    assert isinstance(halfnormal_dist_, numpyro.distributions.HalfNormal)
    assert halfnormal_dist_.scale == 1.5


def test_get_gamma_dist_infer_param():
    gamma_dist_ = gamma_dist(input_vec=jnp.linspace(0, 10, 20))
    assert isinstance(gamma_dist_, numpyro.distributions.Gamma)
    assert gamma_dist_.concentration == 5.0
    assert gamma_dist_.rate == 1.0


def test_get_uniform_dist_error():
    with pytest.raises(ValueError):
        uniform_dist(low=1.0)  # Only low provided without input_vec
    with pytest.raises(ValueError):
        uniform_dist(high=5.0)  # Only high provided without input_vec
    with pytest.raises(ValueError):
        uniform_dist()  # Neither low nor high, and no input_vec


def test_get_gamma_dist_error():
    with pytest.raises(ValueError):
        uniform_dist()  # Neither concentration, nor input_vec


def test_set_fn():
    transformed_fn = set_fn(sample_function)
    result = transformed_fn(2, {"a": 1, "b": 3})
    assert result == 7  # Expected output: 1 + 3*2 = 7


def test_set_kernel_fn():

    # Convert the dummy kernel functions
    new_linear_kernel = set_kernel_fn(linear_kernel_test)
    new_rbf = set_kernel_fn(rbf_test)

    X = jnp.array([[1, 2], [3, 4], [5, 6]])
    Z = jnp.array([[1, 2], [3, 4]])
    params_linear = {"k_scale": 1.0}
    params_rbf = {"k_length": 1.0, "k_scale": 1.0}

    # Assert the transformed function is working correctly
    assert_(jnp.array_equal(linear_kernel_test(X, Z, 1.0), new_linear_kernel(X, Z, params_linear)))
    assert_(jnp.array_equal(rbf_test(X, Z, 1.0, 1.0), new_rbf(X, Z, params_rbf)))


def test_set_kernel_fn_with_jitter():

    jitter = 1e-5

    # Convert the dummy kernel functions
    new_linear_kernel = set_kernel_fn(linear_kernel_test)
    new_rbf = set_kernel_fn(rbf_test)

    X = jnp.array([[1, 2], [3, 4], [5, 6]])
    params_linear = {"k_scale": 1.0}
    params_rbf = {"k_length": 1.0, "k_scale": 1.0}

    # Assert the transformed function is working correctly
    assert_(jnp.array_equal(linear_kernel_test(X, X, 1.0) + jitter * jnp.eye(X.shape[0]), new_linear_kernel(X, X, params_linear, jitter=jitter)))
    assert_(jnp.array_equal(rbf_test(X, X, 1.0, 1.0) + jitter * jnp.eye(X.shape[0]), new_rbf(X, X, params_rbf, jitter=jitter)))


@pytest.mark.parametrize("prior_type", ["normal", "lognormal"])
def test_auto_priors(prior_type):
    prior_fn = auto_priors(sample_function, 1, prior_type, loc=2.0, scale=1.0)
    with numpyro.handlers.seed(rng_seed=1):
        with numpyro.handlers.trace() as tr:
            prior_fn()
    site1 = tr["a"]
    assert_(isinstance(site1['fn'], numpyro.distributions.Distribution))
    assert_equal(site1['fn'].loc, 2.0)
    assert_equal(site1['fn'].scale, 1.0)
    site2 = tr["b"]
    assert_(isinstance(site2['fn'], numpyro.distributions.Distribution))
    assert_equal(site2['fn'].loc, 2.0)
    assert_equal(site2['fn'].scale, 1.0)


@pytest.mark.parametrize("autopriors", [auto_normal_priors, auto_lognormal_priors])
def test_auto_normal_priors(autopriors):
    priors_fn = autopriors(sample_function)
    with numpyro.handlers.seed(rng_seed=1):
        with numpyro.handlers.trace() as tr:
            priors_fn()
    assert_('a' in tr)
    assert_('b' in tr)


@pytest.mark.parametrize("autopriors", [auto_normal_kernel_priors, auto_lognormal_kernel_priors])
def test_auto_normal_kernel_priors(autopriors):
    priors_fn = autopriors(linear_kernel_test)
    with numpyro.handlers.seed(rng_seed=1):
        with numpyro.handlers.trace() as tr:
            priors_fn()
    assert_('k_scale' in tr)


def test_set_noise_kernel_fn():
    from gpax.kernels import RBFKernel

    X = jnp.array([[1, 2], [3, 4], [5, 6]])
    params_i = {"k_length": jnp.array([1.0]), "k_scale": jnp.array(1.0)}
    params = {"k_noise_length": jnp.array([1.0]), "k_noise_scale": jnp.array(1.0)}
    noise_rbf = _set_noise_kernel_fn(RBFKernel)
    assert_(jnp.array_equal(noise_rbf(X, X, params), RBFKernel(X, X, params_i)))
